<?php
/**
 * api: php
 * title: Canonic Autoloader
 * description: Map-based autoloader across .php and .phar resources.
 * version: 0.3.0
 * depends: php:phar, php >= 5.3
 * category: cli
 * license: Public Domain
 *
 *
 * Self-contained autoloader which provides .php and .phar class
 * loading based on a map. Loader and map are also contained in a
 * phar.
 *
 * It's meant for /usr/share/php handling, but can also be utilized
 * for per-project autoloading. It's indifferent to PSR-0/PSR-4 or
 * custom directory schemes. Because of the filename map, identifier
 * names are handled in compliance to PHP semantics (unlike PSR-x);
 * and non-namespaced code works as well. Additionally it prepares
 * function_autoloading support.
 *
 *
 * Using the autoloader is as simple as:
 *
 *   include_once("shared.phar");
 *
 * Invoke it from the commandline to update the internal classmap:
 *
 *   php ./shared.phar
 *
 *
 * Its phar can also be split up. The stub as autoload.php, a separate
 * autoload.update.php script, and an autoload.map.php must then reside
 * within the same directory.
 *
 */


//
if (!class_exists("Canonic_Autoloader")) {
    class Canonic_Autoloader extends ArrayObject {


        // Support for rfc:function_autoloading identifier typing.
        public $types = array(1=>"class", 2=>"function", 4=>"constant");

        // Base directory for relative paths or phar:// paths in map.
        public $basedir = "./";
        
        // Try to rebuild map on missing identifiers.
        public $auto_update = 1;

        

        /**
         * Initialize map.
         *
         * @param  array   $map        Contains a type->identifier->filename map.
         * @param  string  $basedir    Absolute classpath root, usually __DIR__ of shared.phar.
         *
         */
        public function __construct($map, $basedir) {
            $this->basedir = $basedir;
            parent::__construct($map);
        }



        /**
         * Look to-be loaded classname/function up in idmap,
         * and load include script.
         *
         * @param  string  $name       Identifier to autoload.
         * @param  int     $type       Wether it's a class, function, const identifier.
         * @return void
         *
         */
        public function __invoke($name, $type=1) {

            // normalize identifier
            $name = strtolower($name);
            $type = $this->types[$type];
            
            // check for declarations
            if (isset($this[$type][$name]))
            {
                $path = $this[$type][$name];
            }

            // and load according include script
            if (isset($path))
            {
                return include($this->absolute_path($path, $this->basedir));
            }

            // run update?
            if ($this->auto_update-- > 0) {
                if ($map = self::update(__FILE__/*@fixme: not relocatable*/, $this["fp"])) {
                    parent::__construct($map);
                }
                $this($name);
            }
        }
        
        
        
        /**
         * Turn relative paths and phar:// locations into absolute.
         *
         * @param  string  $path       Dirname/Phar path.
         * @param  int     $base       Base directory to prepend on relative paths.
         * @return string
         *
         */
        public function absolute_path($path, $base) {
            // relative phar:
            if (!strncmp($path, "phar://", 7)) {
                if ($path[7] != "/") {
                    $path = "phar://$base/" . substr($path, 7);
                }
            }
            // relative dirname
            elseif ($path[0] != "/") {
                $path = "$base/$path";
            }
            // else it was absolute already
            return $path;
        }




        // File locations, either phar: entries or in ./ local dir. Consider these fixated. In case of a global and local shared.phar, only either declaration of Canonic_Classmap can define them.
        const MAP_FILE = "autoload.map.php";
        const UPDATE_LIB = "autoload.update.php";


        // Create instance and spl_autoload_register it.
        public static function hookup($basefn) {
            spl_autoload_register(
                new Canonic_Autoloader( include(self::pharBase(self::MAP_FILE, $basefn)), dirname($basefn) )
            );
        }

     
        // Recreate classmap within the phar (or as separate autoload.map.php file).
        public static function update($basefn, $fp=NULL) {

            // Load utils manually. Now that's too unfortunate, but necessary for clean .phar distribution and some resiliency.
            include_once(self::pharBase(self::UPDATE_LIB, $basefn));

            // Update from subdirectories.
            return Canonic_Classmap::update( self::pharBase(self::MAP_FILE, $basefn), array(dirname($basefn)), $fp );
        }

        
        // Provide current base directory or phar:// root, depending on if __FILE__ was a .phar or plain php script.
        public static function pharBase($fn="", $basefn) {
            return ( (substr($basefn, -5) == ".phar") ? "phar://" . $basefn : dirname($basefn) ) . "/" . $fn;
        }


    }
}



// Register loader with phar-contained indentifier map.
Canonic_Autoloader::hookup(__FILE__);



// If autoloader.phar is invoked directly from php-cli, trigger update mode.
if (empty($_SERVER["GATEWAY_INTERFACE"]) and isset($_SERVER["argv"][0]) and (basename($_SERVER["argv"][0]) == basename(__FILE__))) {
    Canonic_Autoloader::update(__FILE__);
}



// Phar stub end
__HALT_COMPILER(); ?>
